package bk;

import java.util.ArrayList;

import javax.swing.JTree;

import animation.AlgorithmStage;
import animation.BackwardAction;
import animation.PseudoCodeNode;
import animation.PseudoCodeNode.CodeAction;
import graph.LayeredGraphEdge;
import graph.LayeredGraphNode;

/**
 * the preprocessing stage of the bk node placement algorithm
 * 
 * @author kolja and eren
 *
 */
public class ConflictDetection implements AlgorithmStage {

    private LayeredGraphNode graph;
    private ArrayList<BackwardAction> actions;

    private int i;
    private int l1;
    private int k0;
    private int k1;
    private int l;
    private int k;
    private int hidden_k;
    /**
     * line number in Carstens' pseudocode listing 3.1
     */
    private int pseudo_line;
    private PseudoCodeNode markNode;

    public ConflictDetection(LayeredGraphNode graph) {
        this.graph = graph;
        actions = new ArrayList<>();
        i = 0; // will be increased before first iteration
        l1 = -1; // will be increased before first iteration
        k0 = 0;
        k1 = 0;
        l = 0;
        k = 0;
        hidden_k = 0;
        pseudo_line = 1;
    }

    // TODO: im psuedocode anzeigen
    // TODO: knoten highlighten
    // TODO: rekursion auf subgraphen
    // TODO: backwards actions
    // TODO: work with breakpoints
    @Override
    public StageStatus forwardStep() {
        int old_line = pseudo_line;
        
        switch (pseudo_line) {
        case 1:
            i += 1;
            if (i <= graph.getContainedLayers().size() - 2) {
                pseudo_line += 1;
            } else {
                return StageStatus.FINISHED;
            }
            break;
        case 2:
            k0 = 0;
            l = 0;
            l1 = -1; // will be increased before first iteration
            pseudo_line += 1;
            break;
        case 3:
            l1 += 1;
            if (l1 < graph.getContainedLayers().get(i + 1).size()) {
                pseudo_line += 1;
            } else {
                pseudo_line = 1;
            }
            break;
        case 4:
            if (graph.getContainedLayers().get(i + 1).size() == l1 || incidentToInnerSegmentBetweenLiPlusOneAndLi()) {
                pseudo_line += 1;
            } else {
                pseudo_line = 3;
            }
            break;
        case 5:
            k1 = graph.getContainedLayers().get(i).size() - 1;
            pseudo_line += 1;
            break;
        case 6:
            if (incidentToInnerSegmentBetweenLiPlusOneAndLi()) {
                pseudo_line += 1;
            } else {
                pseudo_line = 9;
            }
            break;
        case 7:
            k1 = graph.getContainedLayers().get(i).indexOf(
                    graph.getContainedLayers().get(i + 1).get(l1).getSortedIncomingEdges().get(0).getSources().get(0));
            pseudo_line = 9;
            break;
        case 9:
            if (l <= l1) {
                pseudo_line += 1;
                // initialize the for loop o next line
                hidden_k -= 1; // because it will be increased next line
            } else {
                pseudo_line = 15;
            }
            break;
        case 10:
            hidden_k += 1;
            if (hidden_k < graph.getContainedLayers().get(i + 1).get(l).getSortedIncomingEdges().size()) {
                k = graph.getContainedLayers().get(i).indexOf(graph.getContainedLayers().get(i + 1).get(l)
                        .getSortedIncomingEdges().get(hidden_k).getSources().get(0));
                pseudo_line += 1;
            } else {
                pseudo_line = 13;
            }
            break;
        case 11:
            hidden_k += 1;
            if (k < k0 || k > k1) {
                graph.getContainedLayers().get(i + 1).get(l).getSortedIncomingEdges().get(hidden_k).setConflicted(true,
                        null);
            }
            pseudo_line = 10;
            break;
        case 13:
            l += 1;
            pseudo_line = 9; 
            break;
        case 15:
            k0 = k1;
            pseudo_line = 3; 
            break;
        }
        
        if (old_line >= 2) {
            graph.getContainedLayers().get(i+1).get(l).setSelected(null);
        }
        
        if (old_line >= 3) {
            graph.getContainedLayers().get(i+1).get(l1).setSelected(null);
        }
        
        if (old_line >= 2) {
            graph.getContainedLayers().get(i).get(k0).setSelected(null);
        }
        
        if (old_line >= 5) {
            graph.getContainedLayers().get(i).get(k1).setSelected(null);
        }
        
        if ((pseudo_line == 10 && old_line == 11) || pseudo_line == 11) {
            graph.getContainedLayers().get(i).get(k).setSelected(null);
        }
        
        System.out.println("going from line " + old_line +" to " + pseudo_line);
        
        return StageStatus.UNFINISHED;
    }

    @Override
    public StageStatus backwardStep() {
        ((PseudoCodeNode) markNode.getParent()).setSelected(true);
        CodeAction action = markNode.setSelected(true);
        boolean breakPoint = action == CodeAction.STOP;
        if (actions.size() == 0) {
            ((PseudoCodeNode) markNode.getParent()).setSelected(false);
            markNode.setSelected(false);
            return StageStatus.FINISHED;
        }
        actions.get(0).reverse();
        actions.remove(0);
        if (breakPoint)
            return StageStatus.BREAKPOINT;
        if (action == CodeAction.SKIP)
            return backwardStep();
        return StageStatus.UNFINISHED;
    }

    @Override
    public PseudoCodeNode createPseudocodeTree(JTree tree) {
        PseudoCodeNode root = new PseudoCodeNode("Preprocessing (mark type 1 conflicts)", tree);
        PseudoCodeNode loopNode = new PseudoCodeNode("Loop through all nodes...", tree);
        markNode = new PseudoCodeNode(
                "If non-inner segment crosses an inner segment whose target is this node, mark the non-inner segment as conflicted",
                tree);
        loopNode.add(markNode);
        root.add(loopNode);
        return root;
    }

    @Override
    public StageStatus forwardStepOver() {
        return forwardStep();
    }

    @Override
    public StageStatus forwardStepOut() {
        StageStatus status = StageStatus.UNFINISHED;
        while (status == StageStatus.UNFINISHED)
            status = forwardStep();
        return status;
    }

    @Override
    public StageStatus backwardStepOver() {
        return backwardStep();
    }

    @Override
    public StageStatus backwardStepOut() {
        StageStatus status = StageStatus.UNFINISHED;
        while (status == StageStatus.UNFINISHED)
            status = backwardStep();
        return status;
    }

    @Override
    public String getDebugString() {
        return "";
    }

    /**
     * part of line 4 and 6 in the pseudocode
     * 
     * @return
     */
    private boolean incidentToInnerSegmentBetweenLiPlusOneAndLi() {
        LayeredGraphNode curr = graph.getContainedLayers().get(i + 1).get(l1);
        for (LayeredGraphEdge e : curr.getIncomingEdges()) {
            if (e.isDummyEdge()) {
                return true;
            }
        }
        return false;
    }
}
